// SPI Slave driver for nRF52840 | gat 26/04/23
#include "SPIS.h"
#include "nrf52840.h"
#include "nrf52840_bitfields.h"
#include <stddef.h> // for NULL
#include "NVIC.h"
#include "GPIO.h"
#include "HSIS.h"
#include "IRQ_Priority.h"

static void dummy(void){}
static void (*Callbacks[])(void) = {dummy, dummy, dummy};
static SPIS_RXD_Type pending_rxd;
static SPIS_TXD_Type pending_txd;
static unsigned char Bool_Update_Pending = 0U;

// -- Utilities --

static NRF_SPIS_Type* GetReg(SPIS_Channel ch)
{
	switch(ch)
	{
		default:
		case SPIS0:
		return NRF_SPIS0;
		case SPIS1:
		return NRF_SPIS1;
		case SPIS2:
		return NRF_SPIS2;
	}
}

static unsigned char GetRegIdx(SPIS_Channel ch)
{
	switch(ch)
	{
		default:
		case SPIS0:
		return 0U;
		case SPIS1:
		return 1U;
		case SPIS2:
		return 2U;
	}
}

// -- Setup --

void SPIS_Setup(SPIS_Config cfg)
{
	NRF_SPIS_Type* reg = GetReg(cfg.Channel);

	// Enable the constant latency sub power mode to minimize the time it takes for the SPIS peripheral to become active after the CSN line is asserted (when the CPU is in sleep mode).
	NRF_POWER->TASKS_CONSTLAT = 1;

	//reg->TASKS_ACQUIRE = 1U;
	//while(!reg->EVENTS_ACQUIRED);
	reg->PSEL.SCK = cfg.CLK_Pin + (cfg.CLK_Port<<5U);
	reg->PSEL.MOSI = cfg.MOSI_Pin + (cfg.MOSI_Port<<5U);
	reg->PSEL.MISO = cfg.MISO_Pin + (cfg.MISO_Port<<5U);
	reg->PSEL.CSN = cfg.CS_Pin + (cfg.CS_Port<<5U);

	reg->TXD.PTR = (uint32_t)NULL;
	reg->TXD.MAXCNT = 0U;
	reg->RXD.PTR = (uint32_t)NULL;
	reg->RXD.MAXCNT = 0U;

	reg->CONFIG = (cfg.Endian) | (cfg.CPHA << SPIS_CONFIG_CPHA_Pos) | (cfg.CPOL << SPIS_CONFIG_CPOL_Pos);
	
	// ORC and DEF ?
	// Event clearing ?

	//reg->SHORTS = SPIS_SHORTS_END_ACQUIRE_Msk;

	IRQ_SET(IRQ_SPI2, ezDAC_SER_PRIO);

	reg->INTENSET = SPIS_INTENSET_ENDRX_Msk | SPIS_INTENSET_END_Msk | SPIS_INTENSET_ACQUIRED_Msk;
	reg->ENABLE = 2;
	reg->TASKS_RELEASE = 1U;
}

// -- Start --

void SPIS_Rx_Callback_Register(SPIS_Channel ch, void (*Callback)(void))
{
	if(Callback != NULL) Callbacks[GetRegIdx(ch)] = Callback;
}

// -- Update --

void SPIS_Update(SPIS_Channel ch, 
	void * tx_ptr, unsigned int tx_size,
	void * rx_ptr, unsigned int rx_size)
{
	NRF_SPIS_Type* reg = GetReg(ch);
	switch(reg->SEMSTAT)
	{
		case SPIS_SEMSTAT_SEMSTAT_SPIS:
		case SPIS_SEMSTAT_SEMSTAT_Free:
			pending_txd.PTR = (unsigned int) tx_ptr;
			pending_txd.MAXCNT = tx_size;
			pending_rxd.PTR = (unsigned int) rx_ptr;
			pending_rxd.MAXCNT = rx_size;
			Bool_Update_Pending = 1U;
			reg->TASKS_ACQUIRE = 1U;
			break;
		case SPIS_SEMSTAT_SEMSTAT_CPU:
			reg->TXD.PTR = (unsigned int) tx_ptr;
			reg->TXD.MAXCNT = tx_size;
			reg->RXD.PTR = (unsigned int) rx_ptr;
			reg->RXD.MAXCNT = rx_size;
			break;
		default:
		case SPIS_SEMSTAT_SEMSTAT_CPUPending: // Eliminate overshoot (we end here if master speaks too fast)
		break;
	}
}

// -- Get Count --

unsigned int SPIS_Get_Rxed_Count(SPIS_Channel ch)
{
	NRF_SPIS_Type* reg = GetReg(ch);
	return reg->RXD.AMOUNT;
}

// -- IRQs --

void Generic_SPIS_IRQ(NRF_SPIS_Type* reg, unsigned char callback_idx)
{
	if(reg->EVENTS_ACQUIRED)
	{
		reg->EVENTS_ACQUIRED = 0U;
		if(Bool_Update_Pending)
		{
			Bool_Update_Pending = 0U;
			reg->TXD.PTR = pending_txd.PTR;
			reg->TXD.MAXCNT = pending_txd.MAXCNT;
			reg->RXD.PTR = pending_rxd.PTR;
			reg->RXD.MAXCNT = pending_rxd.MAXCNT;
		}
		reg->TASKS_RELEASE = 1U;
	}
	if(reg->EVENTS_ENDRX)
	{
		reg->EVENTS_ENDRX = 0U;
	}
	if(reg->EVENTS_END)
	{
		reg->EVENTS_END = 0U;
		//io_tog(DBG_MCU_RX_Port, DBG_MCU_RX_Pin);
		Callbacks[callback_idx]();
	}
	
}
/*
void SPIM0_SPIS0_TWIM0_TWIS0_SPI0_TWI0_IRQHandler(void)
{
	Generic_SPIS_IRQ(NRF_SPIS0, 0U);
}
*/
void SPIM2_SPIS2_SPI2_IRQHandler(void)
{
	
	Generic_SPIS_IRQ(NRF_SPIS2, 2U);
}


